from cv_utils.core import *
from cv_utils.color import *
from cv_utils.video import *
from cv_utils.drawing.matplotlib import *
from cv_utils.drawing.opencv import *
from cv_utils.aruco import *
from cv_utils.colmap import *



import cv_utils.core
import cv_utils.color
import cv2 as cv
import numpy as np
import matplotlib
import matplotlib.pyplot as plt
import pickle
import math
import os
import logging

logger = logging.getLogger(__name__)


CV_ARUCO_DICTIONARIES = {
    cv.aruco.DICT_4X4_50: {'marker_size': 4, 'num_markers': 50, 'april_tag': False},
    cv.aruco.DICT_4X4_100: {'marker_size': 4, 'num_markers': 100, 'april_tag': False},
    cv.aruco.DICT_4X4_250: {'marker_size': 4, 'num_markers': 250, 'april_tag': False},
    cv.aruco.DICT_4X4_1000: {'marker_size': 4, 'num_markers': 1000, 'april_tag': False},
    cv.aruco.DICT_5X5_50: {'marker_size': 5, 'num_markers': 50, 'april_tag': False},
    cv.aruco.DICT_5X5_100: {'marker_size': 5, 'num_markers': 100, 'april_tag': False},
    cv.aruco.DICT_5X5_250: {'marker_size': 5, 'num_markers': 250, 'april_tag': False},
    cv.aruco.DICT_5X5_1000: {'marker_size': 5, 'num_markers': 1000, 'april_tag': False},
    cv.aruco.DICT_6X6_50: {'marker_size': 6, 'num_markers': 50, 'april_tag': False},
    cv.aruco.DICT_6X6_100: {'marker_size': 6, 'num_markers': 100, 'april_tag': False},
    cv.aruco.DICT_6X6_250: {'marker_size': 6, 'num_markers': 250, 'april_tag': False},
    cv.aruco.DICT_6X6_1000: {'marker_size': 6, 'num_markers': 1000, 'april_tag': False},
    cv.aruco.DICT_7X7_50: {'marker_size': 7, 'num_markers': 50, 'april_tag': False},
    cv.aruco.DICT_7X7_100: {'marker_size': 7, 'num_markers': 100, 'april_tag': False},
    cv.aruco.DICT_7X7_250: {'marker_size': 7, 'num_markers': 250, 'april_tag': False},
    cv.aruco.DICT_7X7_1000: {'marker_size': 7, 'num_markers': 1000, 'april_tag': False},
    cv.aruco.DICT_ARUCO_ORIGINAL: {'marker_size': 5, 'num_markers': 1024, 'april_tag': False},
    cv.aruco.DICT_APRILTAG_16h5: {'marker_size': 4, 'num_markers': 30, 'april_tag': True},
    cv.aruco.DICT_APRILTAG_25h9: {'marker_size': 5, 'num_markers': 35, 'april_tag': True},
    cv.aruco.DICT_APRILTAG_36h10: {'marker_size': 6, 'num_markers': 2320, 'april_tag': True},
    cv.aruco.DICT_APRILTAG_36h11: {'marker_size': 6, 'num_markers': 587, 'april_tag': True}
}

CV_CORNER_REFINEMENT_METHODS = {
    'none': cv.aruco.CORNER_REFINE_NONE,
    'subpixel': cv.aruco.CORNER_REFINE_SUBPIX,
    'contour': cv.aruco.CORNER_REFINE_CONTOUR,
    'april_tag': cv.aruco.CORNER_REFINE_APRILTAG
}

class CharucoBoard:

    def __init__(
        self,
        num_squares_x,
        num_squares_y,
        marker_size,
        square_side_length,
        marker_side_length_ratio,
        aruco_dictionary
    ):
        self.num_squares_x = num_squares_x
        self.num_squares_y = num_squares_y
        self.marker_size = marker_size
        self.square_side_length = square_side_length
        self.marker_side_length_ratio = marker_side_length_ratio
        self.marker_side_length = marker_side_length_ratio*square_side_length
        self.aruco_dictionary = aruco_dictionary
        self._cv_charuco_board = cv.aruco.CharucoBoard(
            size=(self.num_squares_x, self.num_squares_y),
            squareLength=self.square_side_length,
            markerLength=self.marker_side_length,
            dictionary=self.aruco_dictionary._cv_aruco_dictionary
        )

    @classmethod
    def from_predefined_dictionary(
        cls,
        num_squares_x=7,
        num_squares_y=5,
        marker_size=6,
        square_side_length=1.0,
        marker_side_length_ratio=0.8,
        april_tag=False
    ):
        logger.info('Creating {}x{} ChArUco board with {}x{} markers, square side length of {}, and marker side length ratio of {} based on pre-defined ArUco dictionary'.format(
            num_squares_x,
            num_squares_y,
            marker_size,
            marker_size,
            square_side_length,
            marker_side_length_ratio
        ))
        aruco_dictionary = ArucoDictionary.get_predefined(
            num_markers=num_squares_x*num_squares_y,
            marker_size=marker_size,
            april_tag=april_tag
        )
        return cls(
            num_squares_x=num_squares_x,
            num_squares_y=num_squares_y,
            marker_size=marker_size,
            square_side_length=square_side_length,
            marker_side_length_ratio=marker_side_length_ratio,
            aruco_dictionary=aruco_dictionary
        )

    @classmethod
    def from_custom_dictionary(
        cls,
        num_squares_x=7,
        num_squares_y=5,
        marker_size=6,
        square_side_length=1.0,
        marker_side_length_ratio=0.8,
        random_seed=0
    ):
        logger.info('Creating {}x{} ChArUco board with {}x{} markers, square side length of {}, and marker side length ratio of {} based on custom ArUco dictionary'.format(
            num_squares_x,
            num_squares_y,
            marker_size,
            marker_size,
            square_side_length,
            marker_side_length_ratio
        ))
        aruco_dictionary = ArucoDictionary.generate_custom(
            num_markers=num_squares_x*num_squares_y,
            marker_size=marker_size,
            base_dictionary=None,
            random_seed=random_seed
        )
        return cls(
            num_squares_x=num_squares_x,
            num_squares_y=num_squares_y,
            marker_size=marker_size,
            square_side_length=square_side_length,
            marker_side_length_ratio=marker_side_length_ratio,
            aruco_dictionary=aruco_dictionary
        )

    def write_image(
        self,
        path,
        image_width_inches=10.5,
        image_height_inches=8,
        image_width_pixels=1000,
        image_height_pixels=None,
        margin_size=0,
        num_border_squares=1
    ):
        image = self.create_image(
            image_width_pixels=image_width_pixels,
            image_height_pixels=image_height_pixels,
            margin_size=margin_size,
            num_border_squares=num_border_squares
        )
        logger.info('Writing image to \'{}\''.format(
            path
        ))
        fig, ax = plt.subplots()
        ax.imshow(
            image,
            cmap = matplotlib.cm.gray,
            interpolation = "nearest"
        )
        ax.axis('off')
        fig.set_size_inches(image_width_inches, image_height_inches)
        plt.savefig(path, facecolor='white', transparent=False)
        plt.close()

    def create_image(
        self,
        image_width_pixels=1000,
        image_height_pixels=None,
        margin_size=0,
        num_border_squares=1
    ):
        image_aspect_ratio = self.num_squares_x/self.num_squares_y
        if image_width_pixels is not None and image_height_pixels is not None:
            pass
        elif image_width_pixels is not None and image_height_pixels is None:
            logger.info('Image width specified as {}. Inferring image height from ChArUco board aspect ratio'.format(
                image_width_pixels
            ))
            image_height_pixels = round(image_width_pixels/image_aspect_ratio)
        elif image_width_pixels is None and image_height_pixels is not None:
            logger.info('Image height specified as {}. Inferring image width from ChArUco board aspect ratio'.format(
                image_height_pixels
            ))
            image_width_pixels = round(image_height_pixels*image_aspect_ratio)
        else:
            raise ValueError('Must specify either image width or image height (or both)')
        image_size_pixels = (image_width_pixels, image_height_pixels)
        logger.info('Creating {}x{} ChArUco board image with {} border squares and margin of {} pixels'.format(
            image_width_pixels,
            image_height_pixels,
            num_border_squares,
            margin_size
        ))
        image = self._cv_charuco_board.generateImage(
            outSize=image_size_pixels,
            marginSize=margin_size,
            borderBits= num_border_squares
        )
        return image

    def calibrate_camera(
        self,
        image_directory,
        min_chessboard_corners=3,
        use_intrinsic_guess=False,
        fix_principal_point=False,
        fix_aspect_ratio=False,
        zero_tangent_distortion=False,
        fix_focal_length=False,
        fix_k1=False,
        fix_k2=False,
        fix_k3=False,
        fix_k4=False,
        fix_k5=False,
        fix_k6=False,
        rational_model=False,
        thin_prism_model=False,
        fix_s1_s2_s3_s4=False,
        tilted_model=False,
        fix_taux_tauy=False,
        camera_matrix_guess=None,
        distortion_coefficients_guess=None,
        calibration_max_iterations=30,
        calibration_accuracy=1e-9,
        min_markers=2,
        marker_corner_refinement_method='none',
        marker_corner_refinement_window_size=5,
        marker_corner_refinement_max_iterations=30,
        marker_corner_refinement_accuracy=0.1,
        marker_detector_parameters=None
    ):
        calibration_termination_criteria = cv_utils.core.termination_criteria(
            max_iterations=calibration_max_iterations,
            accuracy=calibration_accuracy
        )
        calibration_flags, calibration_flag_descriptions = cv_utils.core.camera_calibration_flags(
            use_intrinsic_guess=use_intrinsic_guess,
            fix_principal_point=fix_principal_point,
            fix_aspect_ratio=fix_aspect_ratio,
            zero_tangent_distortion=zero_tangent_distortion,
            fix_focal_length=fix_focal_length,
            fix_k1=fix_k1,
            fix_k2=fix_k2,
            fix_k3=fix_k3,
            fix_k4=fix_k4,
            fix_k5=fix_k5,
            fix_k6=fix_k6,
            rational_model=rational_model,
            thin_prism_model=thin_prism_model,
            fix_s1_s2_s3_s4=fix_s1_s2_s3_s4,
            tilted_model=tilted_model,
            fix_taux_tauy=fix_taux_tauy
        )
        if len(calibration_flag_descriptions) > 0:
            logger.info('Calibrating camera with max iterations {}, accuracy {}, and the following flags: {}'.format(
                calibration_max_iterations,
                calibration_accuracy,
                calibration_flag_descriptions
            ))
        else:
            logger.info('Calibrating camera with max iterations {} and accuracy {}. No calibration flags specified.'.format(
                calibration_max_iterations,
                calibration_accuracy
            ))
        chessboard_corners_object_points = self._cv_charuco_board.getChessboardCorners()
        object_points_array_list = list()
        image_points_array_list = list()
        image_filenames = list()
        image_shape = None
        logger.info('Fetching calibration images from directory \'{}\''.format(
            image_directory
        ))
        with os.scandir(image_directory) as it:
            for directory_entry in it:
                if directory_entry.is_file():
                    logger.info('Fetching calibration image \'{}\''.format(
                        directory_entry.name
                    ))
                    image=cv_utils.core.read_image(directory_entry.path)
                    if image_shape is None:
                        image_shape = image.shape
                    else:
                        if image.shape != image_shape:
                            raise ValueError('Image directory contains images with different shapes')
                    logger.info('Finding ChArUco board corners in image {}'.format(
                        directory_entry.name
                    ))
                    chessboard_corners_dict = self.find_chessboard_corners(
                        image=image,
                        min_markers=min_markers,
                        camera_matrix=None,
                        distortion_coefficients=None,
                        marker_corner_refinement_method=marker_corner_refinement_method,
                        marker_corner_refinement_window_size=marker_corner_refinement_window_size,
                        marker_corner_refinement_max_iterations=marker_corner_refinement_max_iterations,
                        marker_corner_refinement_accuracy=marker_corner_refinement_accuracy,
                        marker_detector_parameters=marker_detector_parameters
                    )
                    if len(chessboard_corners_dict) > min_chessboard_corners:
                        logger.info('Found more than {} corners in {}. Adding to calibration data.'.format(
                            min_chessboard_corners,
                            directory_entry.name
                        ))
                        image_filenames.append(directory_entry.name)
                        object_points_list = list()
                        image_points_list = list()
                        for id, image_point in chessboard_corners_dict.items():
                            object_points_list.append(chessboard_corners_object_points[id])
                            image_points_list.append(image_point)
                        image_points_array = np.stack(image_points_list)
                        object_points_array = np.stack(object_points_list)
                        object_points_array_list.append(object_points_array)
                        image_points_array_list.append(image_points_array)
                    else:
                        logger.info('Found fewer than {} corners. Rejecting image'.format(
                            min_chessboard_corners
                        ))
        logger.info('Calibrating based on ChArUco board corners from {} images'.format(
            len(image_points_array_list)
        ))
        reprojection_error, camera_matrix, distortion_coefficients, rotation_vectors, translation_vectors = cv.calibrateCamera(
            objectPoints=object_points_array_list,
            imagePoints=image_points_array_list,
            imageSize=(image_shape[1], image_shape[0]),
            cameraMatrix=camera_matrix_guess,
            distCoeffs=distortion_coefficients_guess,
            flags=calibration_flags,
            criteria=calibration_termination_criteria,
        )
        extrinsic_parameters = dict()
        for image_filename, rotation_vector, translation_vector in zip(image_filenames, rotation_vectors, translation_vectors):
            extrinsic_parameters[image_filename] = {
                'rotation_vector': rotation_vector,
                'translation_vector': translation_vector
            }
        return camera_matrix, distortion_coefficients, extrinsic_parameters

    def draw_detected_chessboard_corners_opencv(
        self,
        image,
        chessboard_corners_dict,
        corner_color='x00ff00'
    ):
        chessboard_corner_ids_array, chessboard_corners_array = convert_chessboard_corners_dict_to_arrays(
            chessboard_corners_dict
        )
        corner_color_bgr = cv_utils.color.hex_to_bgr(corner_color)
        image = cv.aruco.drawDetectedCornersCharuco(
            image=image,
            charucoCorners=chessboard_corners_array,
            charucoIds=chessboard_corner_ids_array,
            cornerColor=corner_color_bgr
        )
        return image

    def find_chessboard_corners(
        self,
        image,
        min_markers=2,
        camera_matrix=None,
        distortion_coefficients=None,
        refine_markers=True,
        marker_corner_refinement_method='none',
        marker_corner_refinement_window_size=5,
        marker_corner_refinement_max_iterations=30,
        marker_corner_refinement_accuracy=0.1,
        marker_detector_parameters=None
    ):
        logger.info('Finding ChArUco board corners in {}x{} image. Minimum of {} detected markers for each corner. Detecting markers.'.format(
            image.shape[1],
            image.shape[0],
            min_markers
        ))
        charuco_parameters_object = cv.aruco.CharucoParameters()
        setattr(charuco_parameters_object, 'cameraMatrix', camera_matrix)
        setattr(charuco_parameters_object, 'distCoeffs', distortion_coefficients)
        setattr(charuco_parameters_object, 'minMarkers', min_markers)
        setattr(charuco_parameters_object, 'tryRefineMarkers', refine_markers)
        corner_refinement_method_specifier = CV_CORNER_REFINEMENT_METHODS.get(marker_corner_refinement_method)
        if corner_refinement_method_specifier is None:
            raise ValueError('Corner refinement method must be one of the following: {}'.format(
                ', '.join(CV_CORNER_REFINEMENT_METHODS.keys())
            ))
        detector_parameters_object = cv.aruco.DetectorParameters()
        setattr(detector_parameters_object, 'cornerRefinementMethod', corner_refinement_method_specifier)
        setattr(detector_parameters_object, 'cornerRefinementWinSize', marker_corner_refinement_window_size)
        setattr(detector_parameters_object, 'cornerRefinementMaxIterations', marker_corner_refinement_max_iterations)
        setattr(detector_parameters_object, 'cornerRefinementMinAccuracy', marker_corner_refinement_accuracy)
        refine_parameters_object = cv.aruco.RefineParameters()
        charuco_detector = cv.aruco.CharucoDetector(
            board=self._cv_charuco_board,
            charucoParams=charuco_parameters_object,
            detectorParams=detector_parameters_object,
            refineParams=refine_parameters_object,
        )
        chessboard_corners, chessboard_corner_ids, marker_corners, marker_ids = charuco_detector.detectBoard(
            image=image,
        )
        chessboard_corners_dict = convert_chessboard_corner_arrays_to_dict(
            chessboard_corner_ids=chessboard_corner_ids,
            chessboard_corners=chessboard_corners
        )
        if len(chessboard_corners_dict) > 0:
            logger.info('Found {} ChArUco board corners (ids {})'.format(
                len(chessboard_corners_dict),
                ', '.join([str(chessboard_corner_id) for chessboard_corner_id in sorted(chessboard_corners_dict.keys())])
            ))
        else:
            logger.info('Found no ChArUco board corners')
        return chessboard_corners_dict

    def draw_detected_markers_opencv(
        self,
        image,
        marker_corners_dict,
        rejected_image_points,
        border_color='x00ff00'
    ):
        return self.aruco_dictionary.draw_detected_markers_opencv(
            image=image,
            marker_corners_dict=marker_corners_dict,
            rejected_image_points=rejected_image_points,
            border_color=border_color
        )

    def detect_markers(
        self,
        image,
        corner_refinement_method='none',
        corner_refinement_window_size=5,
        corner_refinement_max_iterations=30,
        corner_refinement_accuracy=0.1,
        detector_parameters=None
    ):
        return self.aruco_dictionary.detect_markers(
            image=image,
            corner_refinement_method=corner_refinement_method,
            corner_refinement_window_size=corner_refinement_window_size,
            corner_refinement_max_iterations=corner_refinement_max_iterations,
            corner_refinement_accuracy=corner_refinement_accuracy,
            detector_parameters=detector_parameters
        )


class ArucoDictionary:
    def __init__(
        self,
        num_markers,
        marker_size,
        _cv_aruco_dictionary
    ):
        self.num_markers = num_markers
        self.marker_size = marker_size
        self._cv_aruco_dictionary = _cv_aruco_dictionary

    @classmethod
    def get_predefined(
        cls,
        num_markers=40,
        marker_size=6,
        april_tag=False
    ):
        logger.info('Looking for pre-defined ArUco dictionaries compatible with {} {}x{} markers (AprilTag = {})'.format(
            num_markers,
            marker_size,
            marker_size,
            april_tag
        ))
        compatible_cv_aruco_dictionaries = list(filter(
            lambda item: (
                item[1]['marker_size']==marker_size and
                item[1]['num_markers'] >= num_markers and
                item[1]['april_tag'] == april_tag
            ),
            CV_ARUCO_DICTIONARIES.items()
        ))
        if len(compatible_cv_aruco_dictionaries) == 0:
            raise ValueError('No predefined Aruco dictionaries are compatible with the specified parameters')
        elif len(compatible_cv_aruco_dictionaries) == 1:
            selected_cv_aruco_dictionary = compatible_cv_aruco_dictionaries[0]
            logger.info('Only one compatible pre-defined ArUco dictionary found')
        else:
            selected_cv_aruco_dictionary = min(
                compatible_cv_aruco_dictionaries,
                key=lambda x: x[1]['num_markers']
            )
            logger.info('{} compatible pre-defined ArUco dictionaries found. Selected smallest compatible dictionary ({} markers)'.format(
                len(compatible_cv_aruco_dictionaries),
                selected_cv_aruco_dictionary[1]['num_markers']
            ))
        cv_aruco_dictionary_specifier = selected_cv_aruco_dictionary[0]
        cv_aruco_dictionary = cv.aruco.getPredefinedDictionary(cv_aruco_dictionary_specifier)
        return cls(
            num_markers=num_markers,
            marker_size=marker_size,
            _cv_aruco_dictionary=cv_aruco_dictionary
        )

    @classmethod
    def generate_custom(
        cls,
        num_markers=40,
        marker_size=6,
        base_dictionary=None,
        random_seed=0
    ):
        if base_dictionary is not None:
            logger.info('Generating custom ArUco dictionary with {} {}x{} markers based on dictionary with {} {}x{} markers'.format(
                num_markers,
                marker_size,
                marker_size,
                base_dictionary.num_markers,
                base_dictionary.marker_size,
                base_dictionary.marker_size
            ))
            cv_aruco_dictionary = cv.aruco.extendDictionary(
                nMarkers=num_markers,
                markerSize=marker_size,
                baseDictionary=base_dictionary._cv_aruco_dictionary,
                randomSeed=random_seed
            )
        else:
            logger.info('Generating custom ArUco dictionary with {} {}x{} markers'.format(
                num_markers,
                marker_size,
                marker_size
            ))
            cv_aruco_dictionary = cv.aruco.extendDictionary(
                nMarkers=num_markers,
                markerSize=marker_size,
                randomSeed=random_seed
            )
        return cls(
            num_markers=num_markers,
            marker_size=marker_size,
            _cv_aruco_dictionary=cv_aruco_dictionary
        )

    def bytes_list(self):
        return self._cv_aruco_dictionary.bytesList

    def max_correction_bits(self):
        return self._cv_aruco_dictionary.maxCorrectionBits

    def save(
        self,
        path
    ):
        # Still need to get this working
        raise NotImplementedError()
        # fs = cv.FileStorage(
        #     filename=path,
        #     flags=cv.FILE_STORAGE_WRITE
        # )
        # self._cv_aruco_dictionary.writeDictionary(fs)
        # fs.release()

    @classmethod
    def load(
        cls,
        path
    ):
        # Still need to get this working
        raise NotImplementedError()
        # fs = cv.FileStorage(
        #     filename=path,
        #     flags=cv.FILE_STORAGE_READ
        # )
        # fn = fs.root()
        # cv_aruco_dictionary = cv.aruco.Dictionary.readDictionary(fn)
        # fs.release()
        # return cls(
        #     num_markers=cv_aruco_dictionary.bytesList.shape[0],
        #     marker_size=_cv_aruco_dictionary.markerSize,
        #     _cv_aruco_dictionary=cv_aruco_dictionary
        # )

    def write_images(
        self,
        directory='.',
        filename_prefix='marker',
        filename_extension='png',
        ids=None,
        image_width_inches=10.5,
        image_height_inches=8.0,
        add_label=False,
        image_size_pixels=1000,
        num_border_squares=1
    ):
        num_digits = math.ceil(math.log10(self.num_markers))
        if ids is None:
            ids = range(self.num_markers)
        for id in ids:
            path = os.path.join(
                directory,
                '{{}}_{{:0{}d}}.{{}}'.format(num_digits).format(
                    filename_prefix,
                    id,
                    filename_extension
                )
            )
            self.write_image(
                path=path,
                id=id,
                image_width_inches=image_width_inches,
                image_height_inches=image_height_inches,
                add_label=add_label,
                image_size_pixels=image_size_pixels,
                num_border_squares=num_border_squares
            )

    def write_image(
        self,
        path,
        id,
        image_width_inches=10.5,
        image_height_inches=8.0,
        add_label=True,
        image_size_pixels=1000,
        num_border_squares=1
    ):
        image = self.create_image(
            id=id,
            image_size_pixels=image_size_pixels,
            num_border_squares=num_border_squares
        )
        logger.info('Writing image to \'{}\''.format(
            path
        ))
        fig, ax = plt.subplots()
        ax.imshow(
            image,
            cmap = matplotlib.cm.gray,
            interpolation = "nearest"
        )
        ax.axis('off')
        if add_label:
            ax.set_title(id)
        fig.set_size_inches(image_width_inches, image_height_inches)
        plt.savefig(path, facecolor='white', transparent=False)
        plt.close()

    def create_image(
        self,
        id,
        image_size_pixels=1000,
        num_border_squares=1
    ):
        logger.info('Creating {}x{} image of marker {} with {} border squares'.format(
            image_size_pixels,
            image_size_pixels,
            id,
            num_border_squares
        ))
        image=self._cv_aruco_dictionary.generateImageMarker(
            id=id,
            sidePixels=image_size_pixels,
            borderBits=num_border_squares
        )
        return image

    def estimate_camera_pose_from_detected_markers(
        self,
        marker_corners_dict,
        marker_object_points_dict,
        camera_matrix,
        distortion_coefficients,
        use_extrinsic_guess = False,
        rotation_vector=None,
        translation_vector=None
    ):
        image_points_list = list()
        object_points_list = list()
        for marker_id, object_point in marker_object_points_dict.items():
            if marker_id in marker_corners_dict:
                object_points_list.append(object_point)
                image_points_list.append(marker_corners_dict[marker_id][0])
        if len(object_points_list) < 4:
            raise ValueError('Camera pose estimation requires at least 4 detected markers with known object coordinates. Detected markers: {}. Markers with known object coordinates: {}. Detected markers with known object coordinates: {}'.format(
                len(marker_corners_dict),
                len(marker_object_points_dict),
                len(object_points_list)
            ))
        object_points = np.stack(object_points_list).reshape((-1, 3))
        image_points = np.stack(image_points_list).reshape((-1, 1, 2))
        retval, rotation_vector, translation_vector = cv.solvePnP(
            objectPoints=object_points,
            imagePoints=image_points,
            cameraMatrix=camera_matrix,
            distCoeffs=distortion_coefficients,
            rvec=rotation_vector,
            tvec=translation_vector,
            useExtrinsicGuess=use_extrinsic_guess
        )
        return rotation_vector, translation_vector

    def draw_detected_markers_opencv(
        self,
        image,
        marker_corners_dict,
        rejected_image_points,
        border_color='x00ff00'
    ):
        marker_ids_array, marker_corners_list = convert_marker_corners_dict_to_arrays(
            marker_corners_dict
        )
        border_color_bgr = cv_utils.color.hex_to_bgr(border_color)
        image = cv.aruco.drawDetectedMarkers(
            image=image,
            corners=marker_corners_list,
            ids=marker_ids_array,
            borderColor=border_color_bgr
        )
        return image

    def detect_markers(
        self,
        image,
        corner_refinement_method='none',
        corner_refinement_window_size=5,
        corner_refinement_max_iterations=30,
        corner_refinement_accuracy=0.1,
        detector_parameters=None
    ):
        corner_refinement_method_specifier = CV_CORNER_REFINEMENT_METHODS.get(corner_refinement_method)
        if corner_refinement_method_specifier is None:
            raise ValueError('Corner refinement method must be one of the following: {}'.format(
                ', '.join(CV_CORNER_REFINEMENT_METHODS.keys())
            ))
        detector_parameters_object = cv.aruco.DetectorParameters()
        if detector_parameters is not None:
            for parameter, value in detector_parameters.items():
                setattr(detector_parameters_object, parameter, value)
        setattr(detector_parameters_object, 'cornerRefinementMethod', corner_refinement_method_specifier)
        setattr(detector_parameters_object, 'cornerRefinementWinSize', corner_refinement_window_size)
        setattr(detector_parameters_object, 'cornerRefinementMaxIterations', corner_refinement_max_iterations)
        setattr(detector_parameters_object, 'cornerRefinementMinAccuracy', corner_refinement_accuracy)
        refine_parameters_object = cv.aruco.RefineParameters()
        image_grayscale = cv.cvtColor(image, cv.COLOR_BGR2GRAY)
        logger.info('Detecting ArUco markers with corner refinement method \'{}\', window size {}, max iterations {}, and accuracy {}'.format(
            corner_refinement_method,
            corner_refinement_window_size,
            corner_refinement_max_iterations,
            corner_refinement_accuracy
        ))
        aruco_detector = cv.aruco.ArucoDetector(
            dictionary=self._cv_aruco_dictionary,
            detectorParams=detector_parameters_object,
            refineParams=refine_parameters_object,
        )
        marker_corners, marker_ids, rejected_image_points = aruco_detector.detectMarkers(
            image=image_grayscale,
        )
        marker_corners_dict, rejected_image_points = convert_marker_corner_arrays_to_dict(
            marker_ids=marker_ids,
            marker_corners=marker_corners,
            rejected_image_points=rejected_image_points
        )
        if len(marker_corners_dict) > 0:
            logger.info('Detected {} ArUco markers (ids {}). Rejected {} image points'.format(
                len(marker_corners_dict),
                ', '.join([str(marker_id) for marker_id in sorted(marker_corners_dict.keys())]),
                rejected_image_points.shape[0]
            ))
        else:
            logger.info('Detected no ArUco markers. Rejected {} image points'.format(
                rejected_image_points.shape[0]
            ))
        return marker_corners_dict, rejected_image_points

def convert_marker_corner_arrays_to_dict(
    marker_ids,
    marker_corners,
    rejected_image_points
):
    marker_corners = np.squeeze(np.stack(marker_corners)).reshape((-1, 4, 2))
    marker_ids = np.squeeze(marker_ids).reshape((-1))
    rejected_image_points = np.squeeze(np.stack(rejected_image_points)).reshape((-1, 4, 2))
    if marker_corners.shape[0] != marker_ids.shape[0]:
        raise ValueError('Marker corners array should be same length as marker IDs array but marker corners array has shape {} while marker IDs array has shape {}'.format(
            marker_corners.shape,
            marker_ids.shape
        ))
    if marker_ids.shape[0] > 0:
        marker_ids_list = marker_ids.tolist()
        marker_corners_list = [np.squeeze(corners_array) for corners_array in np.split(marker_corners, marker_corners.shape[0])]
        marker_corners_dict = dict(zip(marker_ids_list, marker_corners_list))
    else:
        marker_corners_dict = dict()
    return marker_corners_dict, rejected_image_points

def convert_marker_corners_dict_to_arrays(
    marker_corners_dict
):
    marker_ids_tuple, marker_corners_tuple = zip(*marker_corners_dict.items())
    marker_ids_array = np.asarray(marker_ids_tuple).reshape(-1, 1)
    marker_corners_list = [marker_corner_array.reshape(1, 4, 2) for marker_corner_array in marker_corners_tuple]
    return marker_ids_array, marker_corners_list

def convert_chessboard_corner_arrays_to_dict(
    chessboard_corner_ids,
    chessboard_corners
):
    chessboard_corners = np.squeeze(chessboard_corners).reshape((-1, 2))
    chessboard_corner_ids = np.squeeze(chessboard_corner_ids).reshape((-1))
    if chessboard_corners.shape[0] != chessboard_corner_ids.shape[0]:
        raise ValueError('Chessboard corners array should be same length as chessboard corner IDs array but chessboard corners array has shape {} while chessboard corner IDs array has shape {}'.format(
            chessboard_corners.shape,
            chessboard_corner_ids.shape
        ))
    if chessboard_corner_ids.shape[0] > 0:
        chessboard_corner_ids_list = chessboard_corner_ids.tolist()
        chessboard_corners_list = [np.squeeze(point_array) for point_array in np.split(chessboard_corners, chessboard_corners.shape[0])]
        chessboard_corners_dict = dict(zip(chessboard_corner_ids_list, chessboard_corners_list))
    else:
        chessboard_corners_dict = dict()
    return chessboard_corners_dict

def convert_chessboard_corners_dict_to_arrays(
    chessboard_corners_dict
):
    chessboard_corner_ids_tuple, chessboard_corners_tuple = zip(*chessboard_corners_dict.items())
    chessboard_corner_ids_array = np.array(chessboard_corner_ids_tuple).reshape((-1, 1))
    chessboard_corners_array = np.array(np.stack(chessboard_corners_tuple)).reshape((-1, 1, 2))
    return chessboard_corner_ids_array, chessboard_corners_array



import cv2 as cv
import numpy as np
import matplotlib.pyplot as plt
import logging

logger = logging.getLogger(__name__)


# Fetch an image from a local image file and return it in OpenCV format
def fetch_image_from_local_drive(image_path):
    image = cv.imread(image_path)
    return image

def plot_2d_image_points(
    image_points,
    image=None,
    image_path=None,
    image_size=None,
    image_alpha=None,
    point_labels=[],
    show_axes=True
):
    image_points = np.asarray(image_points).reshape((-1, 2))
    if image is not None or image_path is not None:
        if image is not None and image_path is not None:
            raise ValueError('Cannot specify both image and image path')
        if image_path is not None:
            image = fetch_image_from_local_drive(image_path)
        if image_size is not None and (image_size[0] != image.shape[1] or image_size[1] != image.shape[0]):
            raise ValueError('Specified image size is ({}, {}) but specified image has size ({}, {})'.format(
                image_size[0],
                image_size[1],
                image.shape[1],
                image.shape[0]
            ))
        image_size = np.array([image.shape[1], image.shape[0]])
        draw_background_image(
            image=image,
            alpha=image_alpha
        )
    draw_2d_image_points(
        image_points,
        point_labels)
    format_2d_image_plot(image_size, show_axes)
    plt.show()

def plot_3d_object_points_topdown(
        object_points,
        room_corners=None,
        point_labels=[]):
    object_points = np.asarray(object_points).reshape((-1, 3))
    draw_3d_object_points_topdown(
        object_points,
        point_labels)
    format_3d_topdown_plot(
        room_corners)
    plt.show()

# Take an image in OpenCV format and plot it as a Matplotlib plot. Calls the
# drawing function above, adds formating, and shows the plot.
def plot_background_image(
        image,
        alpha=None,
        show_axes=True):
    if alpha is None:
        alpha = 0.4
    image_size = np.array([
        image.shape[1],
        image.shape[0]])
    draw_background_image(image, alpha)
    format_2d_image_plot(image_size, show_axes)
    plt.show()

def draw_2d_image_points(
        image_points,
        point_labels=[]):
    image_points = np.asarray(image_points).reshape((-1, 2))
    points_image_u = image_points[:, 0]
    points_image_v = image_points[:, 1]
    plt.plot(
        points_image_u,
        points_image_v,
        '.')
    if len(point_labels) > 0:
        for i in range(len(point_labels)):
            plt.text(points_image_u[i], points_image_v[i], point_labels[i])

def draw_3d_object_points_topdown(
        object_points,
        point_labels=[]):
    object_points = np.asarray(object_points).reshape((-1, 3))
    points_x = object_points[:, 0]
    points_y = object_points[:, 1]
    plt.plot(
        points_x,
        points_y,
        '.')
    if len(point_labels) > 0:
        for i in range(len(point_labels)):
            plt.text(points_x[i], points_y[i], point_labels[i])

# Take an image in OpenCV format and draw it as a background for a Matplotlib
# plot. We separate this from the plotting function below because we might want
# to draw other elements before formatting and showing the chart.
def draw_background_image(
        image,
        alpha=None):
    if alpha is None:
        alpha = 0.4
    plt.imshow(cv.cvtColor(image, cv.COLOR_BGR2RGB), alpha=alpha)

def format_2d_image_plot(
        image_size=None,
        show_axes=True):
    if image_size is not None:
        plt.xlim(0, image_size[0])
        plt.ylim(0, image_size[1])
    if show_axes:
        plt.xlabel(r'$u$')
        plt.ylabel(r'$v$')
        plt.gca().xaxis.set_ticks_position('top')
        plt.gca().xaxis.set_label_position('top')
    else:
        plt.axis('off')
    plt.gca().invert_yaxis()
    plt.gca().set_aspect('equal')

def format_3d_topdown_plot(
        room_corners=None):
    if room_corners is not None:
        plt.xlim(room_corners[0][0], room_corners[1][0])
        plt.ylim(room_corners[0][1], room_corners[1][1])
    plt.xlabel(r'$x$')
    plt.ylabel(r'$y$')
    plt.gca().set_aspect('equal')



import cv2 as cv
import numpy as np
import os
import logging

logger = logging.getLogger(__name__)


def termination_criteria(
    max_iterations=10000,
    accuracy=1e-9
):
    termination_criteria = (cv.TERM_CRITERIA_COUNT & cv.TERM_CRITERIA_EPS, max_iterations, accuracy)
    return termination_criteria

def camera_calibration_flags(
    use_intrinsic_guess=False,
    fix_principal_point=False,
    fix_aspect_ratio=False,
    zero_tangent_distortion=False,
    fix_focal_length=False,
    fix_k1=False,
    fix_k2=False,
    fix_k3=False,
    fix_k4=False,
    fix_k5=False,
    fix_k6=False,
    rational_model=False,
    thin_prism_model=False,
    fix_s1_s2_s3_s4=False,
    tilted_model=False,
    fix_taux_tauy=False,
):
    flags = 0
    flag_descriptions=list()
    if use_intrinsic_guess:
        flags += cv.CALIB_USE_INTRINSIC_GUESS
        flag_descriptions.append('use intrinsic guess')
    if fix_principal_point:
        flags += cv.CALIB_FIX_PRINCIPAL_POINT
        flag_descriptions.append('fix principal point')
    if fix_aspect_ratio:
        flags += cv.CALIB_FIX_ASPECT_RATIO
        flag_descriptions.append('fix aspect ratio')
    if zero_tangent_distortion:
        flags += cv.CALIB_ZERO_TANGENT_DIST
        flag_descriptions.append('zero tangent distortion')
    if fix_focal_length:
        flags += cv.CALIB_FIX_FOCAL_LENGTH
        flag_descriptions.append('fix focal length')
    if fix_k1:
        flags += cv.CALIB_FIX_K1
        flag_descriptions.append('fix k_1')
    if fix_k2:
        flags += cv.CALIB_FIX_K2
        flag_descriptions.append('fix k_2')
    if fix_k3:
        flags += cv.CALIB_FIX_K3
        flag_descriptions.append('fix k_3')
    if fix_k4:
        flags += cv.CALIB_FIX_K4
        flag_descriptions.append('fix k_4')
    if fix_k5:
        flags += cv.CALIB_FIX_K5
        flag_descriptions.append('fix k_5')
    if fix_k6:
        flags += cv.CALIB_FIX_K6
        flag_descriptions.append('fix k_6')
    if rational_model:
        flags += cv.CALIB_RATIONAL_MODEL
        flag_descriptions.append('rational model')
    if thin_prism_model:
        flags += cv.CALIB_THIN_PRISM_MODEL
        flag_descriptions.append('thin prism model')
    if fix_s1_s2_s3_s4:
        flags += cv.CALIB_FIX_S1_S2_S3_S4
        flag_descriptions.append('fix s_1/s_2/s_3/s_4')
    if tilted_model:
        flags += cv.CALIB_TILTED_MODEL
        flag_descriptions.append('tilted model')
    if fix_taux_tauy:
        flags += cv.CALIB_FIX_TAUX_TAUY
        flag_descriptions.append('fix tau_x/tau_y')
    return flags, flag_descriptions

def compose_transformations(
        rotation_vector_1,
        translation_vector_1,
        rotation_vector_2,
        translation_vector_2):
    rotation_vector_1 = np.asarray(rotation_vector_1).reshape(3)
    translation_vector_1 = np.asarray(translation_vector_1).reshape(3)
    rotation_vector_2 = np.asarray(rotation_vector_2).reshape(3)
    translation_vector_2 = np.asarray(translation_vector_2).reshape(3)
    rotation_vector_composed, translation_vector_composed = cv.composeRT(
        rotation_vector_1,
        translation_vector_1,
        rotation_vector_2,
        translation_vector_2)[:2]
    rotation_vector_composed = np.squeeze(rotation_vector_composed)
    translation_vector_composed = np.squeeze(translation_vector_composed)
    return rotation_vector_composed, translation_vector_composed


def invert_transformation(
        rotation_vector,
        translation_vector):
    rotation_vector = np.asarray(rotation_vector).reshape(3)
    translation_vector = np.asarray(translation_vector).reshape(3)
    new_rotation_vector, new_translation_vector = compose_transformations(
        np.array([0.0, 0.0, 0.0]),
        -translation_vector,
        -rotation_vector,
        np.array([0.0, 0.0, 0.0]))
    new_rotation_vector = np.squeeze(new_rotation_vector)
    new_translation_vector = np.squeeze(new_translation_vector)
    return new_rotation_vector, new_translation_vector

def quaternion_vector_to_rotation_vector(quaternion_vector):
    quaternion_vector = np.asarray(quaternion_vector).reshape(4)
    spatial_vector = quaternion_vector[1:]
    qw = quaternion_vector[0]
    spatial_vector_length = np.linalg.norm(spatial_vector)
    unit_vector = spatial_vector/spatial_vector_length
    theta = 2*np.arctan2(spatial_vector_length, qw)
    rotation_vector = theta*unit_vector
    return rotation_vector

def quaternion_vector_to_rotation_matrix(quaternion_vector):
    quaternion_tuple = tuple(np.asarray(quaternion_vector).reshape(4))
    qw, qx, qy, qz = quaternion_tuple
    R = np.array([
        [qw**2 + qx**2 - qy**2 - qz**2, 2*(qx*qy - qw*qz), 2*(qw*qy + qx*qz)],
        [2*(qx*qy + qw*qz), qw**2 - qx**2 + qy**2 - qz**2, 2*(qy*qz - qw*qx)],
        [2*(qx*qz - qw*qy), 2*(qw*qx + qy*qz), qw**2 - qx**2 - qy**2 + qz**2]
    ])
    return R

def rotation_vector_to_rotation_matrix(rotation_vector):
    rotation_vector = np.asarray(rotation_vector).reshape(3)
    rotation_matrix = cv.Rodrigues(rotation_vector)[0]
    return rotation_matrix

def transform_object_points(
        object_points,
        rotation_vector=np.array([0.0, 0.0, 0.0]),
        translation_vector=np.array([0.0, 0.0, 0.0])):
    object_points = np.asarray(object_points)
    rotation_vector = np.asarray(rotation_vector)
    translation_vector = np.asarray(translation_vector)
    if object_points.size == 0:
        return object_points
    object_points = object_points.reshape((-1, 3))
    rotation_vector = rotation_vector.reshape(3)
    translation_vector = translation_vector.reshape(3)
    transformed_points = np.add(
        np.matmul(
            cv.Rodrigues(rotation_vector)[0],
            object_points.T).T,
        translation_vector.reshape((1, 3)))
    transformed_points = np.squeeze(transformed_points)
    return transformed_points


def generate_camera_pose(
        camera_position=np.array([0.0, 0.0, 0.0]),
        yaw=0.0,
        pitch=0.0,
        roll=0.0):
    # yaw: 0.0 points north (along the positive y-axis), positive angles rotate counter-clockwise
    # pitch: 0.0 is level with the ground, positive angles rotate upward
    # roll: 0.0 is level with the ground, positive angles rotate clockwise
    # All angles in radians
    camera_position = np.asarray(camera_position).reshape(3)
    # First: Move the camera to the specified position
    rotation_vector_1 = np.array([0.0, 0.0, 0.0])
    translation_vector_1 = -camera_position
    # Second: Rotate the camera so when we lower to the specified inclination, it will point in the specified compass direction
    rotation_vector_2 = np.array([0.0, 0.0, -(yaw - np.pi / 2)])
    translation_vector_2 = np.array([0.0, 0.0, 0.0])
    # Third: Lower to the specified inclination
    rotation_vector_2_3 = np.array([(np.pi / 2 - pitch), 0.0, 0.0])
    translation_vector_2_3 = np.array([0.0, 0.0, 0.0])
    # Fourth: Roll the camera by the specified angle
    rotation_vector_2_3_4 = np.array([0.0, 0.0, -roll])
    translation_vector_2_3_4 = np.array([0.0, 0.0, 0.0])
    # Combine these four moves
    rotation_vector_1_2, translation_vector_1_2 = compose_transformations(
        rotation_vector_1,
        translation_vector_1,
        rotation_vector_2,
        translation_vector_2)
    rotation_vector_1_2_3, translation_vector_1_2_3 = compose_transformations(
        rotation_vector_1_2,
        translation_vector_1_2,
        rotation_vector_2_3,
        translation_vector_2_3)
    rotation_vector, translation_vector = compose_transformations(
        rotation_vector_1_2_3,
        translation_vector_1_2_3,
        rotation_vector_2_3_4,
        translation_vector_2_3_4)
    rotation_vector = np.squeeze(rotation_vector)
    translation_vector = np.squeeze(translation_vector)
    return rotation_vector, translation_vector


def extract_camera_position(
        rotation_vector,
        translation_vector):
    rotation_vector = np.asarray(rotation_vector).reshape(3)
    translation_vector = np.asarray(translation_vector).reshape(3)
    new_rotation_vector, new_translation_vector = compose_transformations(
        rotation_vector,
        translation_vector,
        -rotation_vector,
        np.array([0.0, 0.0, 0.0]))
    camera_position = -np.squeeze(new_translation_vector)
    return camera_position

def extract_camera_position_rotation_matrix(rotation_matrix, translation_vector):
    rotation_matrix = np.asarray(rotation_matrix).reshape((3,3))
    translation_vector = np.asarray(translation_vector).reshape(3)
    position = np.matmul(rotation_matrix.T, -translation_vector.T)
    return position

def extract_camera_direction(
        rotation_vector,
        translation_vector):
    rotation_vector = np.asarray(rotation_vector).reshape(3)
    translation_vector = np.asarray(translation_vector).reshape(3)
    camera_direction = np.matmul(
        cv.Rodrigues(-rotation_vector)[0],
        np.array([[0.0], [0.0], [1.0]]))
    camera_direction = np.squeeze(camera_direction)
    return camera_direction


def generate_camera_matrix(
        focal_length,
        principal_point):
    focal_length = np.asarray(focal_length).reshape(2)
    principal_point = np.asarray(principal_point).reshape(2)
    camera_matrix = np.array([
        [focal_length[0], 0, principal_point[0]],
        [0, focal_length[1], principal_point[1]],
        [0, 0, 1.0]])
    return camera_matrix


def generate_projection_matrix(
        camera_matrix,
        rotation_vector,
        translation_vector):
    camera_matrix = np.asarray(camera_matrix).reshape((3, 3))
    rotation_vector = np.asarray(rotation_vector).reshape(3)
    translation_vector = np.asarray(translation_vector).reshape(3)
    projection_matrix = np.matmul(
        camera_matrix,
        np.concatenate((
            cv.Rodrigues(rotation_vector)[0],
            translation_vector.reshape((3, 1))),
            axis=1))
    return(projection_matrix)

def ground_grid_camera_view(
    image_width,
    image_height,
    rotation_vector,
    translation_vector,
    camera_matrix,
    distortion_coefficients=np.array([0.0, 0.0, 0.0, 0.0]),
    fill_image=False,
    step=0.1
):
    grid_corners = ground_rectangle_camera_view(
        image_width=image_width,
        image_height=image_height,
        rotation_vector=rotation_vector,
        translation_vector=translation_vector,
        camera_matrix=camera_matrix,
        distortion_coefficients=distortion_coefficients,
        fill_image=fill_image
    )
    grid_points = generate_ground_grid(
        grid_corners=grid_corners,
        step=step
    )
    return grid_points

def ground_rectangle_camera_view(
    image_width,
    image_height,
    rotation_vector,
    translation_vector,
    camera_matrix,
    distortion_coefficients=np.array([0.0, 0.0, 0.0, 0.0]),
    fill_image=False
):
    image_points = np.array([
        [0.0, 0.0],
        [image_width, 0.0],
        [image_width, image_height],
        [0.0, image_height]
    ])
    ground_points=np.empty((4, 3))
    for i in range(4):
        ground_points[i] = ground_point(
            image_point=image_points[i],
            rotation_vector=rotation_vector,
            translation_vector=translation_vector,
            camera_matrix=camera_matrix,
            distortion_coefficients=distortion_coefficients
        )
    x_values_sorted = np.sort(ground_points[:, 0])
    y_values_sorted = np.sort(ground_points[:, 1])
    if fill_image:
        x_min = x_values_sorted[0]
        x_max = x_values_sorted[3]
        y_min = y_values_sorted[0]
        y_max = y_values_sorted[3]
    else:
        x_min = x_values_sorted[1]
        x_max = x_values_sorted[2]
        y_min = y_values_sorted[1]
        y_max = y_values_sorted[2]
    return np.array([
        [x_min, y_min],
        [x_max, y_max]
    ])

def ground_point(
    image_point,
    rotation_vector,
    translation_vector,
    camera_matrix,
    distortion_coefficients=np.array([0.0, 0.0, 0.0, 0.0])
):
    image_point = np.asarray(image_point)
    rotation_vector = np.asarray(rotation_vector)
    translation_vector = np.asarray(translation_vector)
    camera_matrix = np.asarray(camera_matrix)
    distortion_coefficients = np.asarray(distortion_coefficients)
    image_point = image_point.reshape((2))
    rotation_vector = rotation_vector.reshape(3)
    translation_vector = translation_vector.reshape(3)
    camera_matrix = camera_matrix.reshape((3, 3))
    image_point_undistorted = cv.undistortPoints(
        image_point,
        camera_matrix,
        distortion_coefficients,
        P=camera_matrix
    )
    image_point_undistorted = np.squeeze(image_point_undistorted)
    camera_position = np.matmul(
        cv.Rodrigues(-rotation_vector)[0],
        -translation_vector.T
        ).T
    camera_point_homogeneous = np.matmul(
        np.linalg.inv(camera_matrix),
        np.array([image_point_undistorted[0], image_point_undistorted[1], 1.0]).T
    ).T
    camera_direction = np.matmul(
        cv.Rodrigues(-rotation_vector)[0],
        camera_point_homogeneous.T
    ).T
    theta = -camera_position[2]/camera_direction[2]
    ground_point = camera_position + theta*camera_direction
    return ground_point

def generate_ground_grid(
    grid_corners,
    step=0.1
):
    x_grid, y_grid = np.meshgrid(
    np.arange(grid_corners[0, 0], grid_corners[1, 0], step=step),
    np.arange(grid_corners[0, 1], grid_corners[1, 1], step=step)
    )
    grid = np.stack((x_grid, y_grid, np.full_like(x_grid, 0.0)), axis=-1)
    points = grid.reshape((-1, 3))
    return points

def project_points(
    object_points,
    rotation_vector,
    translation_vector,
    camera_matrix,
    distortion_coefficients,
    remove_behind_camera=False,
    remove_outside_frame=False,
    image_corners=None
):
    object_points = np.asarray(object_points).reshape((-1, 3))
    rotation_vector = np.asarray(rotation_vector).reshape(3)
    translation_vector = np.asarray(translation_vector).reshape(3)
    camera_matrix = np.asarray(camera_matrix).reshape((3, 3))
    distortion_coefficients = np.squeeze(np.asarray(distortion_coefficients))
    if object_points.size == 0:
        return np.zeros((0, 2))
    image_points = cv.projectPoints(
        object_points,
        rotation_vector,
        translation_vector,
        camera_matrix,
        distortion_coefficients
    )[0]
    if remove_behind_camera:
        behind_camera_boolean = behind_camera(
            object_points,
            rotation_vector,
            translation_vector
        )
        image_points[behind_camera_boolean] = np.array([np.nan, np.nan])
    if remove_outside_frame:
        outside_frame_boolean = outside_frame(
            object_points,
            rotation_vector,
            translation_vector,
            camera_matrix,
            distortion_coefficients,
            image_corners
        )
        image_points[outside_frame_boolean] = np.array([np.nan, np.nan])
    image_points = np.squeeze(image_points)
    return image_points

def behind_camera(
        object_points,
        rotation_vector,
        translation_vector):
    object_points = np.asarray(object_points)
    rotation_vector = np.asarray(rotation_vector)
    translation_vector = np.asarray(translation_vector)
    if object_points.size == 0:
        return np.zeros((0, 2))
    object_points = object_points.reshape((-1, 3))
    rotation_vector = rotation_vector.reshape(3)
    translation_vector = translation_vector.reshape(3)
    object_points_transformed = transform_object_points(
        object_points,
        rotation_vector,
        translation_vector
    )
    behind_camera_boolean = (object_points_transformed <= 0)[..., 2]
    return behind_camera_boolean

def outside_frame(
    object_points,
    rotation_vector,
    translation_vector,
    camera_matrix,
    distortion_coefficients,
    image_corners
):
    object_points = np.asarray(object_points).reshape((-1, 3))
    rotation_vector = np.asarray(rotation_vector)
    translation_vector = np.asarray(translation_vector).reshape(3)
    camera_matrix = np.asarray(camera_matrix).reshape((3,3))
    distortion_coefficients = np.squeeze(np.asarray(distortion_coefficients))
    image_corners = np.asarray(image_corners).reshape((2,2))
    if object_points.size == 0:
        return np.zeros((0, 2))
    image_points = cv.projectPoints(
        object_points,
        rotation_vector,
        translation_vector,
        camera_matrix,
        np.array([0.0, 0.0, 0.0, 0.0])
    )[0]
    image_points = image_points.reshape((-1, 2))
    outside_frame_boolean = (
        (image_points[:, 0] < image_corners[0, 0]) |
        (image_points[:, 0] > image_corners[1, 0]) |
        (image_points[:, 1] < image_corners[0, 1]) |
        (image_points[:, 1] > image_corners[1, 1])
    )
    return outside_frame_boolean

def undistort_points(
        image_points,
        camera_matrix,
        distortion_coefficients):
    image_points = np.asarray(image_points)
    camera_matrix = np.asarray(camera_matrix)
    distortion_coefficients = np.asarray(distortion_coefficients)
    if image_points.size == 0:
        return image_points
    image_points = image_points.reshape((-1, 1, 2))
    camera_matrix = camera_matrix.reshape((3, 3))
    undistorted_points = cv.undistortPoints(
        image_points,
        camera_matrix,
        distortion_coefficients,
        P=camera_matrix)
    undistorted_points = np.squeeze(undistorted_points)
    return undistorted_points


def estimate_camera_pose_from_image_points(
        image_points_1,
        image_points_2,
        camera_matrix,
        rotation_vector_1=np.array([0.0, 0.0, 0.0]),
        translation_vector_1=np.array([0.0, 0.0, 0.0]),
        distance_between_cameras=1.0):
    image_points_1 = np.asarray(image_points_1)
    image_points_2 = np.asarray(image_points_2)
    camera_matrix = np.asarray(camera_matrix)
    rotation_vector_1 = np.asarray(rotation_vector_1)
    translation_vector_1 = np.asarray(translation_vector_1)
    if image_points_1.size == 0 or image_points_2.size == 0:
        raise ValueError('One or both sets of image points appear to be empty')
    image_points_1 = image_points_1.reshape((-1, 2))
    image_points_2 = image_points_2.reshape((-1, 2))
    if image_points_1.shape != image_points_2.shape:
        raise ValueError('Sets of image points do not appear to be the same shape')
    camera_matrix = camera_matrix.reshape((3, 3))
    rotation_vector_1 = rotation_vector_1.reshape(3)
    translation_vector_1 = translation_vector_1.reshape(3)
    essential_matrix, mask = cv.findEssentialMat(
        image_points_1,
        image_points_2,
        camera_matrix)
    relative_rotation_matrix, relative_translation_vector = cv.recoverPose(
        essential_matrix,
        image_points_1,
        image_points_2,
        camera_matrix,
        mask=mask)[1:3]
    relative_rotation_vector = cv.Rodrigues(relative_rotation_matrix)[0]
    relative_translation_vector = relative_translation_vector * distance_between_cameras
    rotation_vector_2, translation_vector_2 = compose_transformations(
        rotation_vector_1,
        translation_vector_1,
        relative_rotation_vector,
        relative_translation_vector)
    rotation_vector_2 = np.squeeze(rotation_vector_2)
    translation_vector_2 = np.squeeze(translation_vector_2)
    return rotation_vector_2, translation_vector_2


def reconstruct_object_points_from_camera_poses(
        image_points_1,
        image_points_2,
        camera_matrix,
        rotation_vector_1,
        translation_vector_1,
        rotation_vector_2,
        translation_vector_2):
    image_points_1 = np.asarray(image_points_1)
    image_points_2 = np.asarray(image_points_2)
    camera_matrix = np.asarray(camera_matrix)
    rotation_vector_1 = np.asarray(rotation_vector_1)
    translation_vector_1 = np.asarray(translation_vector_1)
    rotation_vector_2 = np.asarray(rotation_vector_2)
    translation_vector_2 = np.asarray(translation_vector_2)
    if image_points_1.size == 0 or image_points_2.size == 0:
        return np.zeros((0, 3))
    image_points_1 = image_points_1.reshape((-1, 2))
    image_points_2 = image_points_2.reshape((-1, 2))
    if image_points_1.shape != image_points_2.shape:
        raise ValueError('Sets of image points do not appear to be the same shape')
    camera_matrix = camera_matrix.reshape((3, 3))
    rotation_vector_1 = rotation_vector_1.reshape(3)
    translation_vector_1 = translation_vector_1.reshape(3)
    rotation_vector_2 = rotation_vector_2.reshape(3)
    translation_vector_2 = translation_vector_2.reshape(3)
    projection_matrix_1 = generate_projection_matrix(
        camera_matrix,
        rotation_vector_1,
        translation_vector_1)
    projection_matrix_2 = generate_projection_matrix(
        camera_matrix,
        rotation_vector_2,
        translation_vector_2)
    object_points_homogeneous = cv.triangulatePoints(
        projection_matrix_1,
        projection_matrix_2,
        image_points_1.T,
        image_points_2.T)
    object_points = cv.convertPointsFromHomogeneous(
        object_points_homogeneous.T)
    object_points = np.squeeze(object_points)
    return object_points


def reconstruct_object_points_from_relative_camera_pose(
        image_points_1,
        image_points_2,
        camera_matrix,
        relative_rotation_vector,
        relative_translation_vector,
        rotation_vector_1=np.array([[0.0], [0.0], [0.0]]),
        translation_vector_1=np.array([[0.0], [0.0], [0.0]]),
        distance_between_cameras=1.0):
    image_points_1 = np.asarray(image_points_1)
    image_points_2 = np.asarray(image_points_2)
    camera_matrix = np.asarray(camera_matrix)
    relative_rotation_vector = np.asarray(relative_rotation_vector)
    relative_translation_vector = np.asarray(relative_translation_vector)
    rotation_vector_1 = np.asarray(rotation_vector_1)
    translation_vector_1 = np.asarray(translation_vector_1)
    if image_points_1.size == 0 or image_points_2.size == 0:
        return np.zeros((0, 3))
    image_points_1 = image_points_1.reshape((-1, 2))
    image_points_2 = image_points_2.reshape((-1, 2))
    if image_points_1.shape != image_points_2.shape:
        raise ValueError('Sets of image points do not appear to be the same shape')
    camera_matrix = camera_matrix.reshape((3, 3))
    relative_rotation_vector = relative_rotation_vector.reshape(3)
    relative_translation_vector = relative_translation_vector.reshape(3)
    rotation_vector_1 = rotation_vector_1.reshape(3)
    translation_vector_1 = translation_vector_1.reshape(3)
    rotation_vector_2, translation_vector_2 = cv.composeRT(
        rotation_vector_1,
        translation_vector_1,
        relative_rotation_vector,
        relative_translation_vector * distance_between_cameras)[:2]
    object_points = reconstruct_object_points_from_camera_poses(
        image_points_1,
        image_points_2,
        camera_matrix,
        rotation_vector_1,
        translation_vector_1,
        rotation_vector_2,
        translation_vector_2)
    return object_points


def reconstruct_object_points_from_image_points(
        image_points_1,
        image_points_2,
        camera_matrix,
        rotation_vector_1=np.array([[0.0], [0.0], [0.0]]),
        translation_vector_1=np.array([[0.0], [0.0], [0.0]]),
        distance_between_cameras=1.0):
    image_points_1 = np.asarray(image_points_1)
    image_points_2 = np.asarray(image_points_2)
    camera_matrix = np.asarray(camera_matrix)
    rotation_vector_1 = np.asarray(rotation_vector_1)
    translation_vector_1 = np.asarray(translation_vector_1)
    if image_points_1.size == 0 or image_points_2.size == 0:
        return np.zeros((0, 3))
    image_points_1 = image_points_1.reshape((-1, 2))
    image_points_2 = image_points_2.reshape((-1, 2))
    if image_points_1.shape != image_points_2.shape:
        raise ValueError('Sets of image points do not appear to be the same shape')
    camera_matrix = camera_matrix.reshape((3, 3))
    rotation_vector_1 = rotation_vector_1.reshape(3)
    translation_vector_1 = translation_vector_1.reshape(3)
    rotation_vector_2, translation_vector_2 = estimate_camera_pose_from_image_points(
        image_points_1,
        image_points_2,
        camera_matrix,
        rotation_vector_1,
        translation_vector_1,
        distance_between_cameras)
    object_points = reconstruct_object_points_from_camera_poses(
        image_points_1,
        image_points_2,
        camera_matrix,
        rotation_vector_1,
        translation_vector_1,
        rotation_vector_2,
        translation_vector_2)
    return object_points

def write_image(
    image,
    path
):
    image_output_directory, image_output_filename = os.path.split(path)
    if image_output_directory is not None:
        os.makedirs(image_output_directory, exist_ok=True)
    cv.imwrite(
        filename=path,
        img=image
    )

def read_image(path):
    image = cv.imread(path)
    return image



import cv_utils.core
import pandas as pd
import numpy as np
import re
import os
import logging

logger = logging.getLogger(__name__)

CALIBRATION_DATA_RE = r'(?P<colmap_image_id>[0-9]+) (?P<qw>[-0-9.]+) (?P<qx>[-0-9.]+) (?P<qy>[-0-9.]+) (?P<qz>[-0-9.]+) (?P<tx>[-0-9.]+) (?P<ty>[-0-9.]+) (?P<tz>[-0-9.]+) (?P<colmap_camera_id>[0-9]+) (?P<image_path>.+)'

def fetch_colmap_output_data_local(
    calibration_directory=None,
    calibration_identifier=None,
    image_data_path=None,
    camera_data_path=None,
    ref_images_data_path=None
):
    """
    Fetches data from COLMAP input and output files and assembles into dataframe.

    The script essentially executes fetch_colmap_image_data_local(),
    fetch_colmap_camera_data_local(), and
    fetch_colmap_reference_image_data_local(); joins their outputs; calculates
    the difference between the camera position inputs and the camera position
    outputs; and assembles everything into a dataframe.

    For details, see documentation for the constituent functions.

    Args:
        calibration_directory (str): Path to directory containing calibrations
        calibration_identifier (str): Identifier for this particular calibration
        image_data_path (str): Explicit path for COLMAP images output file (default is None)
        camera_data_path (str): Explicit path for COLMAP cameras output file (default is None)
        ref_images_data_path (str): Explicit path for COLMAP ref images input file (default is None)

    Returns:
        (DataFrame) Dataframe containing COLMAP output data
    """
    # Fetch COLMAP image output
    df = fetch_colmap_image_data_local(
        calibration_directory=calibration_directory,
        calibration_identifier=calibration_identifier,
        path=image_data_path
    )
    # Fetch COLMAP cameras output
    cameras_df = fetch_colmap_camera_data_local(
        calibration_directory=calibration_directory,
        calibration_identifier=calibration_identifier,
        path=camera_data_path
    )
    df = df.join(cameras_df, on='colmap_camera_id')
    # Fetch COLMAP ref images input
    ref_images_df = fetch_colmap_reference_image_data_local(
        calibration_directory=calibration_directory,
        calibration_identifier=calibration_identifier,
        path=ref_images_data_path
    )
    df = df.join(ref_images_df, on='image_path')
    # Calculate fields
    df['image_path'] = df['image_path'].astype('string')
    df['position_error'] = df['position'] - df['position_input']
    df['position_error_distance'] = df['position_error'].apply(np.linalg.norm)
    return df

def fetch_colmap_image_data_local(
    calibration_directory=None,
    calibration_identifier=None,
    path=None,
):
    """
    Fetches data from COLMAP images output file and assembles into dataframe.

    The script parses the COLMAP images output file, extracting the COLMAP image
    ID, COLMAP camera ID, image path, quaternion vector, and translation vector
    for each image.

    For each image, it then calculates a rotation vector from the quaternion
    vector; calculates a camera position from the rotation vector and
    translation vector; and parses the image path into its subdirectory,
    filename stem, and filename extension.

    By default, the script assumes that the COLMAP images output is in a file
    called images.txt in the directory
    calibration_directory/calibration_identifier. These are the also the default
    path and naming conventions for COLMAP. Alternatively, the user can
    explicitly specify the path for the COLMAP images output file.

    Args:
        calibration_directory (str): Path to directory containing calibrations
        calibration_identifier (str): Identifier for this particular calibration
        path (str): Explicit path for COLMAP image output file (default is None)

    Returns:
        (DataFrame) Dataframe containing image data
    """
    if path is None:
        if calibration_directory is None or calibration_identifier is None:
            raise ValueError('Must specify either image data path or calibration directory and calibration identifier')
        path = os.path.join(
            calibration_directory,
            calibration_identifier,
            'images.txt'
        )
    data_list = list()
    with open(path, 'r') as fp:
        for line in fp.readlines():
            m = re.match(CALIBRATION_DATA_RE, line)
            if m:
                data_list.append({
                    'colmap_image_id': int(m.group('colmap_image_id')),
                    'quaternion_vector': np.asarray([
                        float(m.group('qw')),
                        float(m.group('qx')),
                        float(m.group('qy')),
                        float(m.group('qz'))
                    ]),
                    'translation_vector': np.asarray([
                        float(m.group('tx')),
                        float(m.group('ty')),
                        float(m.group('tz'))
                    ]),
                    'colmap_camera_id': int(m.group('colmap_camera_id')),
                    'image_path': m.group('image_path')

                })
    df = pd.DataFrame(data_list)
    df['rotation_vector'] = df['quaternion_vector'].apply(cv_utils.core.quaternion_vector_to_rotation_vector)
    df['position'] = df.apply(
        lambda row: cv_utils.core.extract_camera_position(
            row['rotation_vector'],
            row['translation_vector']
        ),
        axis=1
    )
    df['image_directory'] = df['image_path'].apply(lambda x: os.path.dirname(os.path.normpath(x))).astype('string')
    df['image_name'] = df['image_path'].apply(lambda x: os.path.splitext(os.path.basename(os.path.normpath(x)))[0]).astype('string')
    df['image_extension'] = df['image_path'].apply(
        lambda x: os.path.splitext(os.path.basename(os.path.normpath(x)))[1][1:]
        if len(os.path.splitext(os.path.basename(os.path.normpath(x)))[1]) > 1
        else None
    ).astype('string')
    df = (
        df
        .reindex(columns=[
            'colmap_image_id',
            'colmap_camera_id',
            'image_path',
            'image_directory',
            'image_name',
            'image_extension',
            'quaternion_vector',
            'rotation_vector',
            'translation_vector',
            'position'
        ])
        .set_index('colmap_image_id')
    )
    return df

def fetch_colmap_camera_data_local(
    calibration_directory=None,
    calibration_identifier=None,
    path=None
):
    """
    Fetches data from COLMAP cameras output file and assembles into dataframe.

    The script parses the COLMAP cameras output file, extracting the COLMAP
    camera ID, COLMAP camera model (e.g., OPENCV), image width, image height,
    and intrinsic calibration parameters for each camera.

    For each camera, it then extracts the camera matrix and distortion
    coefficients from the intrinsic calibration parameters.

    By default, the script assumes that the COLMAP cameras output is in a file
    called cameras.txt in the directory
    calibration_directory/calibration_identifier. These are the also the default
    path and naming conventions for COLMAP. Alternatively, the user can
    explicitly specify the path for the COLMAP cameras output file.

    Args:
        calibration_directory (str): Path to directory containing calibrations
        calibration_identifier (str): Identifier for this particular calibration
        path (str): Explicit path for COLMAP cameras output file (default is
        None)

    Returns:
        (DataFrame) Dataframe containing camera data
    """
    if path is None:
        if calibration_directory is None or calibration_identifier is None:
            raise ValueError('Must specify either camera data path or calibration directory and calibration identifier')
        path = os.path.join(
            calibration_directory,
            calibration_identifier,
            'cameras.txt'
        )
    cameras=list()
    with open(path, 'r') as fp:
        for line_index, line in enumerate(fp):
            if len(line) == 0 or line[0] == '#':
                continue
            word_list = line.split()
            if len(word_list) < 5:
                raise ValueError('Line {} is shorter than expected: {}'.format(
                    line_index,
                    line
                ))
            camera = {
                'colmap_camera_id': int(word_list[0]),
                'colmap_camera_model': word_list[1],
                'image_width': int(word_list[2]),
                'image_height': int(word_list[3]),
                'colmap_parameters': np.asarray([float(parameter_string) for parameter_string in word_list[4:]])
            }
            cameras.append(camera)
    df = pd.DataFrame.from_records(cameras)
    df['camera_matrix'] = df.apply(
        lambda row: colmap_parameters_to_opencv_parameters(
            row['colmap_parameters'],
            row['colmap_camera_model']
        )[0],
        axis=1
    )
    df['distortion_coefficients'] = df.apply(
        lambda row: colmap_parameters_to_opencv_parameters(
            row['colmap_parameters'],
            row['colmap_camera_model']
        )[1],
        axis=1
    )
    df = df.astype({
        'colmap_camera_id': 'int',
        'colmap_camera_model': 'string',
        'image_width': 'int',
        'image_height': 'int',
        'colmap_parameters': 'object',
        'camera_matrix': 'object',
        'distortion_coefficients': 'object'
    })
    df.set_index('colmap_camera_id', inplace=True)
    df = (
        df
        .reindex(columns=[
            'colmap_camera_id',
            'colmap_camera_model',
            'image_width',
            'image_height',
            'colmap_parameters',
            'camera_matrix',
            'distortion_coefficients'
        ])
        .set_index('colmap_camera_id')
    )
    return df

def colmap_parameters_to_opencv_parameters(colmap_parameters, colmap_camera_model):
    if colmap_camera_model == 'SIMPLE_PINHOLE':
        fx = colmap_parameters[0]
        fy = colmap_parameters[0]
        cx = colmap_parameters[1]
        cy = colmap_parameters[2]
        distortion_coefficients = None
    elif colmap_camera_model == 'PINHOLE':
        fx = colmap_parameters[0]
        fy = colmap_parameters[1]
        cx = colmap_parameters[2]
        cy = colmap_parameters[3]
        distortion_coefficients = None
    elif colmap_camera_model == 'SIMPLE_RADIAL':
        fx = colmap_parameters[0]
        fy = colmap_parameters[0]
        cx = colmap_parameters[1]
        cy = colmap_parameters[2]
        distortion_coefficients = np.array([
            colmap_parameters[3],
            0.0,
            0.0,
            0.0
        ])
    elif colmap_camera_model == 'RADIAL':
        fx = colmap_parameters[0]
        fy = colmap_parameters[0]
        cx = colmap_parameters[1]
        cy = colmap_parameters[2]
        distortion_coefficients = np.array([
            colmap_parameters[3],
            colmap_parameters[4],
            0.0,
            0.0
        ])
    elif colmap_camera_model == 'OPENCV':
        fx = colmap_parameters[0]
        fy = colmap_parameters[1]
        cx = colmap_parameters[2]
        cy = colmap_parameters[3]
        distortion_coefficients = np.array([
            colmap_parameters[4],
            colmap_parameters[5],
            colmap_parameters[6],
            colmap_parameters[7]
        ])
    elif colmap_camera_model == 'OPENCV_FISHEYE':
        fx = colmap_parameters[0]
        fy = colmap_parameters[1]
        cx = colmap_parameters[2]
        cy = colmap_parameters[3]
        distortion_coefficients = np.array([
            colmap_parameters[4],
            colmap_parameters[5],
            0.0,
            0.0,
            colmap_parameters[6],
            colmap_parameters[7],
            0.0,
            0.0
        ])
    elif colmap_camera_model == 'FULL_OPENCV':
        fx = colmap_parameters[0]
        fy = colmap_parameters[1]
        cx = colmap_parameters[2]
        cy = colmap_parameters[3]
        distortion_coefficients = np.array([
            colmap_parameters[4],
            colmap_parameters[5],
            colmap_parameters[6],
            colmap_parameters[7],
            colmap_parameters[8],
            colmap_parameters[9],
            colmap_parameters[10],
            colmap_parameters[11]
        ])
    elif colmap_camera_model == 'SIMPLE_RADIAL_FISHEYE':
        fx = colmap_parameters[0]
        fy = colmap_parameters[0]
        cx = colmap_parameters[1]
        cy = colmap_parameters[2]
        distortion_coefficients = np.array([
            colmap_parameters[3],
            0.0,
            0.0,
            0.0
        ])
    elif colmap_camera_model == 'RADIAL_FISHEYE':
        fx = colmap_parameters[0]
        fy = colmap_parameters[0]
        cx = colmap_parameters[1]
        cy = colmap_parameters[2]
        distortion_coefficients = np.array([
            colmap_parameters[3],
            colmap_parameters[4],
            0.0,
            0.0
        ])
    elif colmap_camera_model == 'THIN_PRISM_FISHEYE':
        fx = colmap_parameters[0]
        fy = colmap_parameters[1]
        cx = colmap_parameters[2]
        cy = colmap_parameters[3]
        distortion_coefficients = np.array([
            colmap_parameters[4],
            colmap_parameters[5],
            colmap_parameters[6],
            colmap_parameters[7],
            colmap_parameters[8],
            colmap_parameters[9],
            0.0,
            0.0,
            colmap_parameters[10],
            colmap_parameters[11],
            0.0,
            0.0
        ])
    else:
        raise ValueError('Camera model {} not found'.format(colmap_camera_model))
    camera_matrix = np.array([
        [fx, 0.0, cx],
        [0.0, fy, cy],
        [0.0, 0.0, 1.0]
    ])
    return camera_matrix, distortion_coefficients

def fetch_colmap_reference_image_data_local(
    calibration_directory=None,
    calibration_identifier=None,
    path=None
):
    """
    Fetches data from COLMAP ref images input file and assembles into dataframe.

    The script parses the COLMAP ref images input file, extracting the image
    path and (input) camera position for each image (for comparison with the
    calculated camera position).

    By default, the script assumes that the COLMAP ref images input data is in a
    file called ref_images.txt in the directory
    calibration_directory/calibration_identifier. These are the also the default
    path and naming conventions for COLMAP. Alternatively, the user can
    explicitly specify the path for the COLMAP ref images output file.

    Args:
        calibration_directory (str): Path to directory containing calibrations
        calibration_identifier (str): Identifier for this particular calibration
        path (str): Explicit path for COLMAP ref images input file (default is
        None)

    Returns:
        (DataFrame) Dataframe containing camera position input data
    """
    if path is None:
        if calibration_directory is None or calibration_identifier is None:
            raise ValueError('Must specify either ref image data path or calibration directory and calibration identifier')
        path = os.path.join(
            calibration_directory,
            calibration_identifier,
            'ref_images.txt'
        )
    df = pd.read_csv(
        path,
        header=None,
        delim_whitespace=True,
        names = ['image_path', 'x', 'y', 'z'],
        dtype={
            'image_path': 'string',
            'x': 'float',
            'y': 'float',
            'z': 'float',
        }
    )
    df['position_input'] = df.apply(
        lambda row: np.array([row['x'], row['y'], row['z']]),
        axis=1
    )
    df = (
        df
        .reindex(columns=[
            'image_path',
            'position_input',
        ])
        .set_index('image_path')
    )
    return df




# The MIT License (MIT)
# 
# Copyright (c) 2015-2016 Adrian Rosebrock, http://www.pyimagesearch.com

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
# 
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
# 
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# THIS FILE HAS BEEN MODIFIED FROM THE ORIGINAL BY Benjamin Jaffe-Talberg

from threading import Thread
import sys
import cv2 as cv
import time
import logging

logger = logging.getLogger(__name__)

# import the Queue class from Python 3
if sys.version_info >= (3, 0):
    from queue import Queue

# otherwise, import the Queue class for Python 2.7
else:
    from Queue import Queue


class EagerVideoCapture:
    def __init__(self, path, transform=None, queue_size=64):
        # initialize the file video stream along with the boolean
        # used to indicate if the thread should be stopped or not
        self.capture_object = cv.VideoCapture(path)
        self.stopped = True
        self.exhausted = False
        self.transform = transform

        # initialize the queue used to store frames read from
        # the video file
        self.Q = Queue(maxsize=queue_size)
        # intialize thread
        self.thread = Thread(target=self.update, args=())
        self.thread.daemon = True

    def start(self):
        if self.stopped and not self.exhausted:
            self.stopped = False
            # start a thread to read frames from the file video stream
            self.thread.start()
        return self

    def update(self):
        # keep looping infinitely
        while True:
            # if the thread indicator variable is set, stop the
            # thread
            if self.stopped:
                break

            # otherwise, ensure the queue has room in it
            if not self.Q.full():
                # read the next frame from the file
                (grabbed, frame) = self.capture_object.read()

                # if the `grabbed` boolean is `False`, then we have
                # reached the end of the video file
                if not grabbed:
                    self.exhausted = True
                    self.stopped = True

                # if there are transforms to be done, might as well
                # do them on producer thread before handing back to
                # consumer thread. ie. Usually the producer is so far
                # ahead of consumer that we have time to spare.
                #
                # Python is not parallel but the transform operations
                # are usually OpenCV native so release the GIL.
                #
                # Really just trying to avoid spinning up additional
                # native threads and overheads of additional
                # producer/consumer queues since this one was generally
                # idle grabbing frames.
                if self.transform:
                    frame = self.transform(frame)

                # add the frame to the queue
                self.Q.put((grabbed, frame))
            else:
                time.sleep(0.1)  # Rest for 10ms, we have a full queue

        self.capture_object.release()

    def read(self):
        logger.debug("Reading from EagerVideoCapture queue, current size: {}".format(self.Q.qsize()))
        # return next frame in the queue
        return self.Q.get()

    # Insufficient to have consumer use while(more()) which does
    # not take into account if the producer has reached end of
    # file stream.
    def running(self):
        return self.more() or not self.stopped

    def more(self):
        # return True if there are still frames in the queue. If stream is not stopped, try to wait a moment
        tries = 0
        while self.Q.qsize() == 0 and not self.stopped and tries < 5:
            time.sleep(0.1)
            tries += 1

        return self.Q.qsize() > 0

    def stop(self):
        # indicate that the thread should be stopped
        self.stopped = True
        # wait until stream resources are released (producer thread might be still grabbing frame)
        self.thread.join()



import cv2 as cv
import cv_utils.color
import pandas as pd
import numpy as np
import math
import logging

logger = logging.getLogger(__name__)


MARKER_DICT = {
    '+': cv.MARKER_CROSS,
    'x': cv.MARKER_TILTED_CROSS,
    '*': cv.MARKER_STAR,
    'd': cv.MARKER_DIAMOND,
    's': cv.MARKER_SQUARE,
    '^': cv.MARKER_TRIANGLE_UP,
    'v': cv.MARKER_TRIANGLE_DOWN
}


def draw_circle(
    original_image,
    coordinates,
    radius=6,
    line_width=1.5,
    color='#00ff00',
    fill=True,
    alpha=1.0
):
    center = tuple(map(lambda x: int(round(x)), coordinates))
    for coordinate in center:
        if abs(coordinate) > 2**30:
            return original_image
    color_bgr = cv_utils.color.hex_to_bgr(color)
    thickness = math.ceil(line_width)
    if fill:
        thickness = cv.FILLED
    overlay_image = original_image.copy()
    # print(center)
    overlay_image = cv.circle(
        img=overlay_image,
        center=center,
        radius=math.ceil(radius),
        color=color_bgr,
        thickness=thickness,
        lineType=cv.LINE_AA
    )
    new_image = cv.addWeighted(
        overlay_image,
        alpha,
        original_image,
        1 - alpha,
        0
    )
    return new_image


def draw_line(
    original_image,
    coordinates,
    line_width=1.5,
    color='#00ff00',
    alpha=1.0
):
    pt1 = tuple(map(lambda x: int(round(x)), coordinates[0]))
    pt2 = tuple(map(lambda x: int(round(x)), coordinates[1]))
    for coordinate in pt1:
        if abs(coordinate) > 2**30:
            return original_image
    for coordinate in pt2:
        if abs(coordinate) > 2**30:
            return original_image
    color_bgr = cv_utils.color.hex_to_bgr(color)
    overlay_image = original_image.copy()
    overlay_image = cv.line(
        img=overlay_image,
        pt1=pt1,
        pt2=pt2,
        color=color_bgr,
        thickness=math.ceil(line_width),
        lineType=cv.LINE_AA
    )
    new_image = cv.addWeighted(
        overlay_image,
        alpha,
        original_image,
        1 - alpha,
        0
    )
    return new_image

def draw_timestamp(
    original_image,
    timestamp,
    padding=5,
    font_face=cv.FONT_HERSHEY_PLAIN,
    font_scale=1.5,
    text_line_width=1,
    text_color='#ffffff',
    text_alpha=1.0,
    box_line_width=0,
    box_color='#000000',
    box_fill=True,
    box_alpha=0.3
):

    image_height, image_width, image_depth = original_image.shape
    upper_right_coordinates = [image_width - padding, padding]
    timestamp_text = pd.to_datetime(timestamp, utc=True).strftime('%Y-%m-%dT%H:%M:%S.%fUTC')
    new_image = draw_text_box(
        original_image=original_image,
        anchor_coordinates=upper_right_coordinates,
        text=timestamp_text,
        horizontal_alignment='right',
        vertical_alignment='top',
        font_face=font_face,
        font_scale=font_scale,
        text_line_width=text_line_width,
        text_color=text_color,
        text_alpha=text_alpha,
        box_line_width=box_line_width,
        box_color=box_color,
        box_fill=box_fill,
        box_alpha=box_alpha
    )
    return new_image

def draw_text_box(
    original_image,
    anchor_coordinates,
    text,
    horizontal_alignment='center',
    vertical_alignment='bottom',
    font_face=cv.FONT_HERSHEY_PLAIN,
    font_scale=1.0,
    text_line_width=1,
    text_color='#00ff00',
    text_alpha=1.0,
    box_line_width=1.5,
    box_color='#00ff00',
    box_fill=True,
    box_alpha=1.0
):
    text_box_coordinates = get_text_box_coordinates(
        anchor_coordinates=anchor_coordinates,
        text=text,
        horizontal_alignment=horizontal_alignment,
        vertical_alignment=vertical_alignment,
        font_face=font_face,
        font_scale=font_scale,
        line_width=text_line_width
    )
    new_image = draw_rectangle(
        original_image=original_image,
        coordinates=text_box_coordinates,
        line_width=box_line_width,
        color=box_color,
        fill=box_fill,
        alpha=box_alpha
    )
    new_image = draw_text(
        original_image=new_image,
        anchor_coordinates=anchor_coordinates,
        text=text,
        horizontal_alignment=horizontal_alignment,
        vertical_alignment=vertical_alignment,
        font_face=font_face,
        font_scale=font_scale,
        line_width=text_line_width,
        color=text_color,
        alpha=text_alpha
    )
    return new_image

def draw_text(
    original_image,
    anchor_coordinates,
    text,
    horizontal_alignment='center',
    vertical_alignment='bottom',
    font_face=cv.FONT_HERSHEY_PLAIN,
    font_scale=1.0,
    line_width=1,
    color='#00ff00',
    alpha=1.0
):
    for anchor_coordinate in anchor_coordinates:
        if abs(anchor_coordinate) > 2**30:
            return original_image
    color_bgr = cv_utils.color.hex_to_bgr(color)
    thickness = math.ceil(line_width)
    text_org_u, text_org_v = get_text_org(
        anchor_coordinates=anchor_coordinates,
        text=text,
        horizontal_alignment=horizontal_alignment,
        vertical_alignment=vertical_alignment,
        font_face=font_face,
        font_scale=font_scale,
        line_width=line_width
    )
    overlay_image = original_image.copy()
    overlay_image = cv.putText(
        img=overlay_image,
        text=text,
        org=(int(round(text_org_u)), int(round(text_org_v))),
        fontFace=font_face,
        fontScale=font_scale,
        color=color_bgr,
        thickness=thickness,
        lineType=cv.LINE_AA
    )
    new_image = cv.addWeighted(
        overlay_image,
        alpha,
        original_image,
        1 - alpha,
        0
    )
    return new_image

def get_text_box_coordinates(
    anchor_coordinates,
    text,
    horizontal_alignment,
    vertical_alignment,
    font_face=cv.FONT_HERSHEY_PLAIN,
    font_scale=1.0,
    line_width=1
):
    text_org_u, text_org_v = get_text_org(
        anchor_coordinates=anchor_coordinates,
        text=text,
        horizontal_alignment=horizontal_alignment,
        vertical_alignment=vertical_alignment,
        font_face=font_face,
        font_scale=font_scale,
        line_width=line_width
    )
    thickness = math.ceil(line_width)
    text_size, baseline = cv.getTextSize(
        text=text,
        fontFace=font_face,
        fontScale=font_scale,
        thickness=thickness
    )
    text_width, text_height = text_size
    baseline += thickness
    text_box_lower_left_u = text_org_u
    text_box_lower_left_v  = text_org_v + baseline/2
    text_box_upper_right_u = text_org_u + text_width
    text_box_upper_right_v = text_org_v - text_height - baseline/2
    coordinates = (
        (text_box_lower_left_u, text_box_lower_left_v),
        (text_box_upper_right_u, text_box_upper_right_v)
    )
    return coordinates

def get_text_org(
    anchor_coordinates,
    text,
    horizontal_alignment,
    vertical_alignment,
    font_face=cv.FONT_HERSHEY_PLAIN,
    font_scale=1.0,
    line_width=1
):
    thickness = math.ceil(line_width)
    text_size, baseline = cv.getTextSize(
        text=text,
        fontFace=font_face,
        fontScale=font_scale,
        thickness=thickness
    )
    text_width, text_height = text_size
    baseline += thickness
    anchor_coordinates_u, anchor_coordinates_v = anchor_coordinates
    if horizontal_alignment == 'left':
        text_org_u = anchor_coordinates_u
    elif horizontal_alignment == 'center':
        text_org_u = anchor_coordinates_u - text_width / 2
    elif horizontal_alignment == 'right':
        text_org_u = anchor_coordinates_u - text_width
    else:
        raise ValueError('Horizontal aligment \'{}\' not recognized'.format(horizontal_alignment))
    if vertical_alignment == 'top':
        text_org_v = anchor_coordinates_v + text_height
    elif vertical_alignment == 'middle':
        text_org_v = anchor_coordinates_v + text_height / 2
    elif vertical_alignment == 'bottom':
        text_org_v = anchor_coordinates_v
    else:
        raise ValueError('Vertical aligment \'{}\' not recognized'.format(vertical_alignment))
    return (text_org_u, text_org_v)

def draw_point(
    original_image,
    coordinates,
    marker='.',
    marker_size=10,
    line_width=1,
    color='#00ff00',
    alpha=1.0
):
    position = tuple(map(lambda x: int(round(x)), coordinates))
    for coordinate in position:
        if abs(coordinate) > 2**30:
            return original_image
    if marker == '.':
        return draw_circle(
            original_image,
            coordinates,
            radius=marker_size / 2,
            color=color,
            fill=True,
            alpha=alpha
        )
    color_bgr = cv_utils.color.hex_to_bgr(color)
    markerType = MARKER_DICT.get(marker)
    overlay_image = original_image.copy()
    overlay_image = cv.drawMarker(
        img=overlay_image,
        position=position,
        color=color_bgr,
        markerType=markerType,
        markerSize=math.ceil(marker_size),
        thickness=math.ceil(line_width),
        line_type=cv.LINE_AA
    )
    new_image = cv.addWeighted(
        overlay_image,
        alpha,
        original_image,
        1 - alpha,
        0
    )
    return new_image


def draw_rectangle(
    original_image,
    coordinates,
    line_width=1.5,
    color='#00ff00',
    fill=True,
    alpha=1.0
):
    pt1 = tuple(map(lambda x: int(round(x)), coordinates[0]))
    pt2 = tuple(map(lambda x: int(round(x)), coordinates[1]))
    for coordinate in pt1:
        if abs(coordinate) > 2**30:
            return original_image
    for coordinate in pt2:
        if abs(coordinate) > 2**30:
            return original_image
    color_bgr = cv_utils.color.hex_to_bgr(color)
    thickness = math.ceil(line_width)
    if fill:
        thickness = cv.FILLED
    overlay_image = original_image.copy()
    overlay_image = cv.rectangle(
        img=overlay_image,
        pt1=pt1,
        pt2=pt2,
        color=color_bgr,
        thickness=thickness,
        lineType=cv.LINE_AA
    )
    new_image = cv.addWeighted(
        overlay_image,
        alpha,
        original_image,
        1 - alpha,
        0
    )
    return new_image

def draw_polygon(
    original_image,
    vertices,
    color='#00ff00',
    alpha=1.0
):
    vertices = np.asarray(vertices)
    num_vertices = vertices.shape[0]
    if vertices.shape != (num_vertices, 2):
        raise ValueError('Vertices must be of shape (num_vertices, 2)')
    for vertex_index in range(num_vertices):
        for coordinate_index in range(2):
            vertices[vertex_index, coordinate_index] = int(round(vertices[vertex_index, coordinate_index]))
    vertices = vertices.astype(int)
    color_bgr = cv_utils.color.hex_to_bgr(color)
    overlay_image = original_image.copy()
    overlay_image = cv.fillConvexPoly(
        img=overlay_image,
        points=vertices,
        color=color_bgr,
        lineType=cv.LINE_AA,
        shift=0
    )
    new_image = cv.addWeighted(
        overlay_image,
        alpha,
        original_image,
        1 - alpha,
        0
    )
    return new_image



import logging

logger = logging.getLogger(__name__)

def hex_to_rgb(color_hex):
    red, green, blue = bytes.fromhex(color_hex.lstrip('#'))
    color_rgb = (red, green, blue)
    return color_rgb


def rgb_to_bgr(color_rgb):
    color_bgr = (color_rgb[2], color_rgb[1], color_rgb[0])
    return color_bgr


def hex_to_bgr(color_hex):
    return rgb_to_bgr(hex_to_rgb(color_hex))



import os
from setuptools import setup, find_packages

BASEDIR = os.path.dirname(os.path.abspath(__file__))
VERSION = open(os.path.join(BASEDIR, 'VERSION')).read().strip()

BASE_DEPENDENCIES = [
    'opencv-python>=4.5.1',
    'pandas>=1.2.2',
    'numpy>=1.20.1',
    'matplotlib>=3.3.4'
]

# allow setup.py to be run from any path
os.chdir(os.path.normpath(BASEDIR))

setup(
    name='wf-cv-utils',
    packages=find_packages(),
    version=VERSION,
    include_package_data=True,
    description='Miscellaneous utilities for working with camera data',
    long_description=open('README.md').read(),
    url='https://github.com/WildflowerSchools/wf-cv-utils',
    author='Theodore Quinn',
    author_email='ted.quinn@wildflowerschools.org',
    install_requires=BASE_DEPENDENCIES,
    keywords=['cv'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
    ]
)



import cv_utils.core
import cv_utils.eager_video_capture as vc_queue
import cv2 as cv
import pandas as pd
import datetime
import os
import logging

logger = logging.getLogger(__name__)


class VideoInput:
    def __init__(
        self,
        input_path,
        start_time=None,
        queue_frames=True,
        queue_size=64
    ):
        if not os.path.isfile(input_path):
            raise ValueError('No file at specified path: {}'.format(input_path))

        self.queue_frames = queue_frames

        if self.queue_frames:
            self.video_queue = vc_queue.EagerVideoCapture(input_path, queue_size=queue_size)
            self.capture_object = self.video_queue.capture_object
        else:
            self.video_queue = None
            self.capture_object = cv.VideoCapture(input_path)

        self.video_parameters = VideoParameters(
            start_time=start_time,
            frame_width=self.capture_object.get(cv.CAP_PROP_FRAME_WIDTH),
            frame_height=self.capture_object.get(cv.CAP_PROP_FRAME_HEIGHT),
            fps=self.capture_object.get(cv.CAP_PROP_FPS),
            frame_count=self.capture_object.get(cv.CAP_PROP_FRAME_COUNT),
            fourcc_int=self.capture_object.get(cv.CAP_PROP_FOURCC)
        )

    def is_opened(self):
        return self.capture_object.isOpened()

    def close(self):
        self.capture_object.release()

    def write_frame_by_timestamp(
        self,
        timestamp,
        path
    ):
        image=self.get_frame_by_timestamp(timestamp)
        cv_utils.core.write_image(
            image=image,
            path=path
        )

    def write_frame_by_frame_number(
        self,
        frame_number,
        path
    ):
        image = self.get_frame_by_frame_number(frame_number)
        cv_utils.core.write_image(
            image=image,
            path=path
        )

    def write_frame_by_milliseconds(
        self,
        milliseconds,
        path
    ):
        image = self.get_frame_by_milliseconds(milliseconds)
        cv_utils.core.write_image(
            image=image,
            path=path
        )

    def get_frame_by_timestamp(self, timestamp):
        if self.video_parameters.start_time is None or self.video_parameters.fps is None or self.video_parameters.frame_count is None:
            raise ValueError('Valid video start time, FPS, and frame count required to get frame by timestamp')
        try:
            timestamp = pd.to_datetime(timestamp, utc=True).to_pydatetime()
        except Exception as e:
            raise ValueError('Cannot parse start time: {}'.format(timestamp))
        frame_number = round(
            (timestamp - self.video_parameters.start_time).total_seconds()*
            self.video_parameters.fps
        )
        logger.debug('Target timestamp is {}. Selected frame number {} at timestamp {}'.format(
            timestamp.isoformat(),
            frame_number,
            (self.video_parameters.start_time + datetime.timedelta(seconds=frame_number/self.video_parameters.fps)).isoformat()
        ))
        if frame_number < 0 or frame_number > self.video_parameters.frame_count:
            raise ValueError('Specified datetime is outside the time range of the video')
        return self.get_frame_by_frame_number(frame_number)

    def get_frame_by_frame_number(self, frame_number):
        self.capture_object.set(cv.CAP_PROP_POS_FRAMES, frame_number)
        return self.get_frame()

    def get_frame_by_milliseconds(self, milliseconds):
        self.capture_object.set(cv.CAP_PROP_POS_MSEC, milliseconds)
        return self.get_frame()

    def get_frame(self):
        if self.queue_frames:
            # Wait to start queue until the first frame is needed. This is in case
            # additional VideoCaptureProperties need to be applied before the
            # video begins streaming (i.e. cv.CAP_PROP_POS_FRAMES - start frame)
            self.video_queue.start()
            ret, frame = self.video_queue.read()
        else:
            ret, frame = self.capture_object.read()

        if ret:
            return frame
        else:
            return None


class VideoOutput:
    def __init__(
        self,
        output_path,
        video_parameters
    ):
        self.video_parameters = video_parameters
        self.writer_object = cv.VideoWriter(
            output_path,
            fourcc=self.video_parameters.fourcc_int,
            fps=self.video_parameters.fps,
            frameSize=(
                self.video_parameters.frame_width,
                self.video_parameters.frame_height
            )
        )

    def is_opened(self):
        return self.writer_object.isOpened()

    def close(self):
        self.writer_object.release()

    def write_frame(self, frame):
        self.writer_object.write(frame)


class VideoParameters:
    def __init__(
        self,
        start_time=None,
        frame_width=None,
        frame_height=None,
        fps=None,
        frame_count=None,
        fourcc_int=None
    ):
        self.start_time = None
        self.frame_width = None
        self.frame_height = None
        self.fps = None
        self.frame_count = None
        self.fourcc_int = None
        self.time_index = None
        if start_time is not None:
            try:
                self.start_time = pd.to_datetime(start_time, utc=True).to_pydatetime()
            except Exception as e:
                raise ValueError('Cannot parse start time: {}'.format(start_time))
            # try:
            #     self.start_time = start_time.astimezone(datetime.timezone.utc)
            # except Exception as e:
            #     try:
            #         self.start_time = datetime.fromisoformat(start_time).astimezone(datetime.timezone.utc)
                # except Exception as e:
                #     raise ValueError('Cannot parse start time: {}'.format(start_time))
        if frame_width is not None:
            try:
                self.frame_width = int(frame_width)
            except Exception as e:
                raise ValueError('Frame width must be convertible to integer')
        if frame_height is not None:
            try:
                self.frame_height = int(frame_height)
            except Exception as e:
                raise ValueError('Frame height must be convertible to integer')
        if fps is not None:
            try:
                self.fps = float(fps)
            except Exception as e:
                raise ValueError('FPS must be convertible to float')
        if frame_count is not None:
            try:
                self.frame_count = int(frame_count)
            except Exception as e:
                raise ValueError('Frame count must be convertible to integer')
        if fourcc_int is not None:
            try:
                self.fourcc_int = int(fourcc_int)
            except Exception as e:
                raise ValueError('FourCC code must be convertible to integer')


def fourcc_string_to_int(fourcc_string):
    return cv.VideoWriter_fourcc(*fourcc_string)


def fourcc_int_to_string(fourcc_int):
    return "".join([chr((int(fourcc_int) >> 8 * i) & 0xFF) for i in range(4)])
